"""Provides the tools to build classes corresponding to XML elements
in GPX and TCX files.
"""

import datetime

from lxml import etree

from . import util

def create_data_prop(path, conv_type=int):
  """Add property inside an ActivityElement class definition that accesses data
  using :meth:`~ActivityElement.get_data`.

  Args:
    path (str): The tag name or path of the desired subelement of the
      contained lxml element whose text will be retrieved. Text will be
      retrieved from the first matching subelement.
    conv_type (data type): Data type that the element text will
      be converted to. Python type, or datetime.datetime to read a time using
      :meth:`dateutil.parser.isoparse`. Defaults to ``int``.
  Returns:
    property: accessor for the underlying lxml element's data.

  Examples:

    >>> class MyElement(ActivityElement):
    ...   my_prop = create_data_prop('Time', datetime.datetime)
    ...   \"\"\"Property docstring goes here (if desired).\"\"\"
    ...   [...]

  """
  return property(
    lambda self: self.get_data(path, conv_type=conv_type), 
    # doc=f':obj:`{conv_type.__name__}`'
  )


def create_attr_prop(key, conv_type=int):
  """Add property inside an ActivityElement class definition that accesses data
  using :meth:`~ActivityElement.get_attr`.

  Args:
    key (str): The attribute name within the contained lxml element.
    conv_type (data type): Data type that the element text will
      be converted to. Python type, or datetime.datetime to read a time using
      :meth:`dateutil.parser.isoparse`. Defaults to ``int``.
  Returns:
    property: accessor for underlying lxml element's attribute data.

  Examples:

    >>> class MyElement(ActivityElement):
    ...   my_prop = create_attr_prop('Time', datetime.datetime)
    ...   \"\"\"Property docstring goes here (if desired)\"\"\"
    ...   [...]

  """
  return property(
    lambda self: self.get_attr(key, conv_type=conv_type),
    # doc=f':obj:`{conv_type.__name__}`'
  )


def create_descendent_prop(descendent_class):
  """Add descendent list accessor property inside an ActivityElement class
  definition.

  The property returns all the current element's descendents matching the
  tag name defined in the descendent class.

  Args:
    descendent_class (ActivityElement): Must be a subclass of :class:`ActivityElement`.
  Returns:
    property: accessor for underlying lxml element's descendent list.

  Examples:

    >>> class MyElement(ActivityElement):
    ...   my_prop = create_descendent_prop(MySubElement)
    ...   \"\"\"Property docstring goes here (if desired)\"\"\"
    ...   [...]

  """
  return property(
    lambda self: [
      descendent_class(e) 
      for e in self.elem.xpath(f'.//{descendent_class.TAG}')
    ],
    doc=f':obj:`list` of :class:`{descendent_class.__name__}`: '
    # f'All descendents of the contained lxml element with tag name '
    f'All element descendents with tag name "{descendent_class.TAG}".',
  )


class ActivityElement(object):
  """Base class for creating compositions with :class:`~lxml.etree._Element`.

  TCX and GPX files are both XML documents that follow their own particular
  schema. ActivityElement subclasses provide access to the underlying XML
  elements' data, attributes, descendent elements.
  
  Args:
    lxml_elem (:class:`~lxml.etree._Element`): XML element object that has
      been read in by the :mod:`lxml.etree` package. The tag name of the
      element must match the :attr:`TAG` attribute of this class, or else
      a TypeError will be raised.
  """

  TAG = 'element'
  """XML tag name of the element. 
  
  If an instance of the class is initialized from a 
  :class:`lxml.etree._Element` with any other tag name, a TypeError will
  be raised.
  """

  def __init__(self, lxml_elem):
    if not isinstance(lxml_elem, etree._Element):
      raise TypeError(
        f'Expected lxml element, not {type(lxml_elem).__name__}'
      )
    
    if lxml_elem.tag != self.TAG:
      raise ValueError(
        f'Expected lxml element with "{self.TAG}" tag, not "{lxml_elem.tag}".'
      )

    self.elem = lxml_elem

  @classmethod
  def _add_data_properties(cls, **property_paths_types):
    """Add properties to cls that access specific paths using :meth:`get_data`.
    
    Args:
      cls: Class to add the properties to.
      **property_paths_types: kwargs mapping property_name : tuple(path, data_type)

        - property_name will be added to cls.
        - path is the tag name or path of the desired subelement of the 
          contained lxml element whose text will be retrieved
        - data_type is the type that the subelement's text will be converted
          to. A python type, or datetime.datetime to read a timestamp using 
          :meth:`dateutil.parser.isoparse`.

    Examples:

      >>> class MyElement(ActivityElement):
      ...   [...]
      >>> MyElement._add_data_properties(time=('Time', datetime.datetime))

    See also:
      `pandas.core.accessor.PandasDelegate._add_delegate_accessors <https://github.com/pandas-dev/pandas/blob/v1.2.4/pandas/core/accessor.py#L57-L108>`_

    """
    for prop_name, (path, data_type) in property_paths_types.items():
      f = create_data_prop(path, conv_type=data_type)
      setattr(cls, prop_name, f)

  @classmethod
  def _add_attr_properties(cls, **property_keys_types):
    """Add properties to cls that access specific keys using :meth:`get_attr`.
    
    Args:
      cls: Class to add the properties to.
      **property_keys_types: kwargs mapping property_name : tuple(key, data_type)

        - property_name will be added to cls.
        - key is the attribute name within the contained lxml element whose 
          text will be retrieved. 
        - data_type is the type that the subelement's text will be converted
          to. A python type, or datetime.datetime to read a timestamp using 
          :meth:`dateutil.parser.isoparse`.

    Examples:

      >>> class MyElement(ActivityElement):
      ...   [...]
      >>> MyElement._add_attr_properties(start_time=('StartTime', datetime.datetime))
    """
    for prop_name, (key, data_type) in property_keys_types.items():
      f = create_attr_prop(key, conv_type=data_type)
      setattr(cls, prop_name, f)

  @classmethod
  def _add_descendent_properties(cls, **descendent_classes):
    """Add properties to cls that return all descendents matching a path.
    
    Args:
      cls: Class to add the properties to.

      **descendent_classes: kwargs mapping property_name : descendent class.
    
        - property_name will be added to cls.
        - All values must be a subclass of :class:`ActivityElement`.

    Examples:

      >>> class MyElement(ActivityElement):
      ...   [...]
      >>> MyElement._add_descendent_properties(subelements=MySubElement)
    """
    for prop_name, descendent_class in descendent_classes.items():
      f = create_descendent_prop(descendent_class)
      setattr(cls, prop_name, f)
      # cls.add_descendent_list_property(prop_name, descendent_class)    

  def get_data(self, path, conv_type=str):
    """Retrieve data using the contained lxml element's 
    :meth:`~lxml.etree._Element.findtext` and convert its type.
    
    TODO: 
      * Consider if this wants to be a delegated ``findtext`` method, which
        could be accomplished by moving the type conversion out of this method.

    Args:
      path (str): The tag name or path of the desired subelement of the
        contained lxml element whose text will be retrieved. Text will be
        retrieved from the first matching subelement.
      conv_type (data type): Data type that the element text will
        be converted to. Python type, or datetime.datetime to read a time using
        :meth:`dateutil.parser.isoparse`. Defaults to ``str``.

    Returns:
      The retrieved data in the requested type, or None if no subelements
      match ``path``.

    Examples:
      Text can be converted to another Python type:

      >>> e_lat = ActivityElement(etree.fromstring(
      ...   '<element><Position><LatitudeDegrees>40.0</LatitudeDegrees></Position></element>'
      ... ))
      >>> e_lat.get_data('Position/LatitudeDegrees', conv_type='float')
      40.0
      
      Or a timestamp can be read in:

      >>> e_time = ActivityElement(etree.fromstring(
      ...   '<element><Time>2021-02-26T19:51:08.000Z</Time></element>'
      ... ))
      >>> e_time.get_data('Time', conv_type=datetime.datetime)
      datetime.datetime(2021, 2, 26, 19, 51, 8, tzinfo=tzutc())

    """
    data = self.elem.findtext(path)

    if data is None:
      return None

    conv_func = util.get_conv_func(conv_type)

    return conv_func(data)

  def get_attr(self, key, conv_type=str):
    """Retrieve data using contained lxml element's 
    :meth:`~lxml.etree._Element.get` and convert its type.

    Args:
      key (str): The attribute name within the contained lxml element.
      conv_type (data type): Data type that the attribute text will
        be converted to. Python type, or datetime.datetime to read a time using
        :meth:`dateutil.parser.isoparse`. Defaults to ``str``.

    Returns:
      The retrieved attribute data in the requested type, or None if 
      the contained lxml element does not have an attribute named ``key``.

    Examples:

      >>> e = ActivityElement(etree.fromstring(
      ...   '<element StartTime="2021-02-26T19:51:07.000Z"></element>
      ... ))
      >>> e.get_attr('StartTime', conv_type=datetime.datetime)
      datetime.datetime(2021, 2, 26, 19, 51, 7, tzinfo=tzutc())

    """
    attr = self.elem.get(key)

    if attr is None:
      return None

    conv_func = util.get_conv_func(conv_type)

    return conv_func(attr)


def add_xml_data(**property_paths_types):
  """Add properties to a class that access data using :meth:`get_data`.

  This provides an alternative usage to directly calling 
  :meth:`ActivityElement._add_data_properties` below a class definition.

  Args:
    **property_paths_types: kwargs mapping property_name : tuple(path, data_type)

    - property_name will be added to the decorated class.
    - path is the tag name or path of the desired subelement of the 
      contained lxml element whose text will be retrieved
    - data_type is the type that the subelement's text will be converted
      to. A python type, or datetime.datetime to read a timestamp using 
      :meth:`dateutil.parser.isoparse`.
  Returns:
    callable: A class decorator.
  
  Examples:
  
    >>> @add_xml_data(time=('Time', datetime.datetime))
    ... class MyElement(ActivityElement):
    ...   [...]

  See also:
    :meth:`ActivityElement._add_data_properties`
  """

  def add_data_properties(cls):
    cls._add_data_properties(**property_paths_types)
    return cls

  return add_data_properties


def add_xml_attr(**property_keys_types):
  """Add properties to a class that access data using :meth:`get_attr`.

  This provides an alternative usage to directly calling 
  :meth:`ActivityElement._add_attr_properties` below a class definition.

  Args:
    **property_keys_types: kwargs mapping property_name : tuple(key, data_type)

      - property_name will be added to the decorated class.
      - key is the attribute name within the contained lxml element whose 
        text will be retrieved. 
      - data_type is the type that the subelement's text will be converted
        to. A python type, or datetime.datetime to read a timestamp using 
        :meth:`dateutil.parser.isoparse`.
  Returns:
    callable: A class decorator.
  
  Examples:

    >>> @add_xml_attr(start_time=('StartTime', datetime.datetime))
    ... class MyElement(ActivityElement):
    ...   [...]

  See also:
    :meth:`ActivityElement._add_attr_properties`
  """
  def add_attr_properties(cls):
    cls._add_attr_properties(**property_keys_types)
    return cls

  return add_attr_properties


def add_xml_descendents(**descendent_classes):
  """Add properties to a class that return all descendents matching a path.

  This provides an alternative usage to directly calling 
  :meth:`ActivityElement._add_descendent_properties` below a class definition.

  Args:
    **descendent_classes: kwargs mapping property_name : descendent class.
    
      - Each key will be a property name added to the decorated class.
      - All values must be a subclass of :class:`ActivityElement`.
  Returns:
    callable: A class decorator 

  Examples:

    >>> @add_xml_descendents(subelements=MySubElement)
    ... class MyElement(ActivityElement):
    ...   [...]

  See also:
    :meth:`ActivityElement._add_descendent_properties`
  """
  def add_descendent_properties(cls):
    cls._add_descendent_properties(**descendent_classes)
    return cls

  return add_descendent_properties



import datetime

from dateutil import parser
from lxml import objectify


def get_conv_func(conv_type):
  if conv_type == datetime.datetime:
    return parser.isoparse
  
  # Assume this is a Python type
  return conv_type


def get_time(time_text):
  """Returns a tz-aware datetime."""
  try:
    return parser.isoparse(time_text)
  except TypeError:
    return None


def strip_namespaces(element):
  """Strip namespaces from an elements to permit easier operations.

  https://stackoverflow.com/questions/18159221/remove-namespace-and-prefix-from-xml-in-python-using-lxml
  
  From what I've heard, lxml isn't really smart enough to deal with
  them better than I can (by hand). I only have a few file types to
  deal with. 
  
  Kind of like how FitParse can deal with profile.xlsx, but I've opted
  to do that by hand. I think this is analogous?
  """
  for elem in element.getiterator():
    # Guard against comments.
    if not hasattr(elem.tag, 'find'): continue
    
    # Locate the extent of the namespace text and remove it.
    i = elem.tag.find('}')
    if i >= 0:
      elem.tag = elem.tag[i+1:]

  # Get rid of all the `'ns5': 'http://...'` and `xsi:type` business
  objectify.deannotate(element, cleanup_namespaces=True)


# -*- coding: utf-8 -*-
""".gpx file reader architecture.

See also:
  
  `GPX schema <https://www.topografix.com/GPX/1/1/>`_
    Official documentation for GPX elements and file structure.

  `Garmin's GPX trackpoint extension schema <https://www8.garmin.com/xmlschemas/TrackPointExtensionv2.xsd>`_
    XML file describing the schema for Garmin's additional GPX trackpoint elements.
"""
import datetime
import io

from lxml import etree

from . import util
from .base import (
  ActivityElement,
  add_xml_data, add_xml_attr, add_xml_descendents, 
  create_data_prop, create_attr_prop, create_descendent_prop
)


class Trackpoint(ActivityElement):
  """Represents a single data sample corresponding to a point in time.
  
  The most granular of data contained in the file.
  """
  TAG = 'trkpt'

  time = create_data_prop('time', datetime.datetime)
  """datetime.datetime: Timestamp when trackpoint was recorded.
  
  See also:
    :ref:`data.timestamp`
  """

  altitude_m = create_data_prop('ele', float)
  """float: Elevation of ground surface in meters above sea level."""

  hr = create_data_prop('extensions/TrackPointExtension/hr', int)
  """int: Heart rate."""

  cadence_rpm = create_data_prop('extensions/TrackPointExtension/cad', int)
  """int: Cadence in RPM.
  
  See also:
    :ref:`data.cadence`
  """

  lat = create_attr_prop('lat', float)
  """float: Latitude in degrees N (-90 to 90)."""

  lon = create_attr_prop('lon', float)
  """float: Longitude in degrees E (-180 to 180)."""


class Segment(ActivityElement):
  """Holds a list of trackpoints which are logically connected in order.
  
  To represent a single GPS track where GPS reception was lost, or the 
  GPS receiver was turned off, start a new Track Segment for each continuous
  span of track data.
  """
  TAG = 'trkseg'

  trackpoints = create_descendent_prop(Trackpoint)


@add_xml_data(
  name=('name', str),
  activity_type=('type', str),
)
class Track(ActivityElement):
  """An ordered list of trackpoints describing a path."""
  TAG = 'trk'

  segments = create_descendent_prop(Segment)
  trackpoints = create_descendent_prop(Trackpoint)


@add_xml_data(name=('metadata/name', str))
@add_xml_attr(
  creator=('creator', str),
  version=('version', str),
)
class Gpx(ActivityElement):
  """Represents an entire .gpx file object."""

  TAG = 'gpx'

  @classmethod
  def from_file(cls, file_obj):
    """Initialize a Gpx element from a file-like object.

    Args:
      file_obj (str, bytes, io.StringIO, io.BytesIO): File-like object. 
        If str, either filename or a string representation of XML 
        object. If str or StringIO, the encoding should not be declared
        within the string.

    Returns:
      Gpx: An instance initialized with the :class:`~lxml.etree._Element`
      that was read in.

    See also:
      https://lxml.de/tutorial.html#the-parse-function

    """
    if not isinstance(file_obj, (str, bytes, io.StringIO, io.BytesIO)):
      raise TypeError(f'file object type not accepted: {type(file_obj)}')

    if isinstance(file_obj, str) and not file_obj.lower().endswith('.gpx'):
      file_obj = io.StringIO(file_obj)
    elif isinstance(file_obj, bytes):
      file_obj = io.BytesIO(file_obj)

    tree = etree.parse(file_obj)
    root = tree.getroot()

    util.strip_namespaces(root)

    return cls(root)

  start_time = create_data_prop('metadata/time', datetime.datetime)
  """datetime.datetime: Timestamp at start of recording.
  
  See also:
    :ref:`data.timestamp`
  """

  # @property
  # def name(self):
  #   # GPX elements sometimes don't have their own name, so default to
  #   # the first track's name.
  #   return self.tracks[0].name

  tracks = create_descendent_prop(Track)
  segments = create_descendent_prop(Segment)
  trackpoints = create_descendent_prop(Trackpoint)



# -*- coding: utf-8 -*-
""".tcx file reader architecture.

Originated in `heartandsole <https://github.com/aaron-schroeder/heartandsole/blob/affc028c266e108e669d93a99b19dbb8e176db49/heartandsole/filereaders.py#L295>`_
back in the day.

See also:

  `Garmin's TCX schema <https://www8.garmin.com/xmlschemas/TrainingCenterDatabasev2.xsd>`_
    XML file describing the schema for TCX files.

  `Garmin's ActivityExtension schema <https://www8.garmin.com/xmlschemas/ActivityExtensionv2.xsd>`_
    XML file describing Garmin's extensions to the TCX schema.
"""
import datetime
import io

from dateutil import tz
from lxml import etree

from . import util
from .base import (
  ActivityElement, 
  add_xml_data, add_xml_attr, add_xml_descendents,
  # add_data_props, add_attr_props, add_descendent_props,
  create_data_prop, create_attr_prop, create_descendent_prop
)


class Trackpoint(ActivityElement):
  """Represents a single data sample corresponding to a point in time.
  
  The most granular of data contained in the file.
  """
  TAG = 'Trackpoint'

  time = create_data_prop('Time', datetime.datetime)
  """datetime.datetime: Timestamp when trackpoint was recorded.
  
  See also:
    :ref:`data.timestamp`
  """
  
  # OR
  # @data_prop('dummy_tag', float)
  # 

  lat = create_data_prop('Position/LatitudeDegrees', float)
  """float: Latitude in degrees N (-90 to 90)."""

  lon = create_data_prop('Position/LongitudeDegrees', float)
  """float: Longitude in degrees E (-180 to 180)."""

  altitude_m = create_data_prop('AltitudeMeters', float)
  """float: Elevation of ground surface in meters above sea level."""

  distance_m = create_data_prop('DistanceMeters', float)
  """float: Cumulative distance from the start of the activity, in meters.
  
  See also:
    :ref:`data.distance`
  """

  hr = create_data_prop('HeartRateBpm/Value', int)
  """int: Heart rate."""

  speed_ms = create_data_prop('Extensions/TPX/Speed', float)
  """float: Speed in meters per second.
  
  TODO:
    * Consider looping this in to the explanation for distance data,
      maybe it can be a whole thing on sensor fusion with links.
  """

  cadence_rpm = create_data_prop('Extensions/TPX/RunCadence', int)
  """int: Cadence in RPM.
  
  See also:
    :ref:`data.cadence`
  """


class Track(ActivityElement):
  """In a running TCX file, there is typically one Track per Lap.

  As far as I can tell, in a running file, Tracks and Laps are one
  and the same; when a Lap starts or ends, so does its contained Track.
  
  Made up of 1 or more :class:`Trackpoint` in xml file.
  """
  TAG = 'Track'

  trackpoints = create_descendent_prop(Trackpoint)


@add_xml_data(
  intensity=('Intensity', str),
  trigger_method=('TriggerMethod', str),
)
class Lap(ActivityElement):
  """Represents one bout from {start/lap} -> {lap/stop}.

  There is at least one lap per activity file, created by the `start` button
  press and ended by the `stop` button press. Hitting the `lap` button begins
  a new lap. Hitting the pause button stops data recording, but the same lap
  resumes after the pause.
  
  Made up of 1 or more :class:`Track` in the XML structure.
  """
  TAG = 'Lap'

  start_time = create_attr_prop('StartTime', datetime.datetime)
  """datetime.datetime: Timestamp of lap start.
  
  See also:
    :ref:`data.timestamp`
  """

  total_time_s = create_data_prop('TotalTimeSeconds', float)
  """float: Total lap time, in seconds, as reported by the device.
  
  This is timer time, not elapsed time; it does not include any time when
  the device is paused.

  See also:
    :ref:`data.tcx.start_stop_pause`
  """

  distance_m = create_data_prop('DistanceMeters', float)
  """float: Total lap distance, in meters, as reported by the device.
  
  See also:
    :ref:`data.distance`
  """

  max_speed_ms = create_data_prop('MaximumSpeed', float)
  """float: The maximum speed achieved during the lap as reported by the 
  device, in meters per second.
  """

  avg_speed_ms = create_data_prop('Extensions/LX/AvgSpeed', float)
  """float: The average speed during the lap as reported by the device,
  in meters per second.
  """

  hr_avg = create_data_prop('AverageHeartRateBpm/Value', int)
  """float: average heart rate during the lap as reported by the device."""

  hr_max = create_data_prop('MaximumHeartRateBpm/Value', int)
  """float: maximum heart rate during the lap as reported by the device."""

  cadence_avg = create_data_prop('Extensions/LX/AvgRunCadence', int)
  """float: average cadence during the lap as reported by the device, in RPM.
  
  See also:
    :ref:`data.cadence`
  """

  cadence_max = create_data_prop('Extensions/LX/MaxRunCadence', int)
  """float: maximum cadence during the lap as reported by the device, in RPM.
  
  See also:
    :ref:`data.cadence`
  """

  calories = create_data_prop('Calories', int)
  """float: Calories burned during the lap as approximated by the device."""

  tracks = create_descendent_prop(Track)
  trackpoints = create_descendent_prop(Trackpoint)


@add_xml_data(product_id=('Creator/ProductID', int))
class Activity(ActivityElement):
  """TCX files representing a run should only contain one Activity.

  Contains one or more :class:`Lap` elements.
  """
  TAG = 'Activity'

  start_time = create_data_prop('Id', datetime.datetime)
  """datetime.datetime: Timestamp for activity start time.
  
  See also:
    :ref:`data.timestamp`
  """

  sport = create_attr_prop('Sport', str)
  """str: Activity sport. 
  
  Restricted to "Running", "Biking", or "Other" according to Garmin's TCX 
  file schema.
  """

  device = create_data_prop('Creator/Name', str)
  """str: Device brand name."""

  device_id = create_data_prop('Creator/UnitId', int)
  """int: Device ID - specific to the individual device."""
  
  @property
  def version(self):
    major = self.get_data('Creator/Version/VersionMajor', int)
    minor = self.get_data('Creator/Version/VersionMinor', int)
    return f'{major}.{minor}'

  @property
  def build(self):
    major = self.get_data('Creator/Version/BuildMajor', int)
    minor = self.get_data('Creator/Version/BuildMinor', int)
    return f'{major}.{minor}'

  laps = create_descendent_prop(Lap)
  tracks = create_descendent_prop(Track)
  trackpoints = create_descendent_prop(Trackpoint)


@add_xml_data(
  creator=('Author/Name', str),
  part_number=('Author/PartNumber', str)
)
class Tcx(ActivityElement):
  """Represents an entire .tcx file object."""

  TAG = 'TrainingCenterDatabase'

  @classmethod
  def from_file(cls, file_obj):
    """Initialize a Tcx element from a file-like object.

    Args:
      file_obj (str, bytes, io.StringIO, io.BytesIO): File-like object. 
        If str, either filename or a string representation of XML 
        object. If str or StringIO, the encoding should not be declared
        within the string.

    Returns:
      Tcx: An instance initialized with the :class:`~lxml.etree._Element`
      that was read in.

    See also:
      https://lxml.de/tutorial.html#the-parse-function

    TODO:
      * Consider if the master file readers should use a separate for
        creation, akin to etree.parse or from_string. Gah. I could make
        it more flexible, sure. But it seems weird to just have one
        random subclass with a different init. Maybe Tcx is the only
        one who gets this special extra class. Also naturally am
        thinking about delegating it.
      * How many delegated methods are we looking at now? parse, from_string,
        find_text, get, xpath, ... others?
        
    """
    if not isinstance(file_obj, (str, bytes, io.StringIO, io.BytesIO)):
      raise TypeError(f'file object type not accepted: {type(file_obj)}')

    if isinstance(file_obj, str) and not file_obj.lower().endswith('.tcx'):
      file_obj = io.StringIO(file_obj)
    elif isinstance(file_obj, bytes):
      file_obj = io.BytesIO(file_obj)

    # Note: tree is an ElementTree, which is just a thin wrapper
    # around root, which is an element
    tree = etree.parse(file_obj)
    root = tree.getroot()
    util.strip_namespaces(root)

    return cls(root)

  # Below here are convenience properties that access data from
  # descendent elements. Not sure if they all stay.

  @property
  def device(self):
    return self.activities[0].device

  @property
  def distance_m(self):
    return sum([lap.distance_m for lap in self.laps])

  @property
  def calories(self):
    return sum([lap.calories for lap in self.laps])

  @property
  def lap_time_s(self):
    return sum([lap.total_time_s for lap in self.laps])

  @property
  def num_laps(self):
    return len(self.laps)

  @property
  def num_bouts(self):
    return len(self.tracks)

  @property
  def num_records(self):
    return len(self.trackpoints)

  activities = create_descendent_prop(Activity)
  laps = create_descendent_prop(Lap)
  tracks = create_descendent_prop(Track)
  trackpoints = create_descendent_prop(Trackpoint)
